/* client.c */

#include <sys/utsname.h>
#include <net/ethernet.h>
#include <sys/ioctl.h>
#include "fumied.h"
#include "udpipgen.h"

int (*Fsm[MAX_STATES])();		/* finite state machine */
int CurrState;				/* current state */
int PrevState;				/* previous state */

time_t ReqSentTime;		/* time when FUMIEDREQUEST message is sent */

struct ifinfo Ifbuf;

int fumiedSocket;
udpipMessage UdpIpMsg;
unsigned char ClientHwAddr[ETHER_ADDR_LEN];

struct fumiedMessage *FumiedMsgRecv;
struct fumiedMessage *FumiedMsgSend = (struct fumiedMessage *)&UdpIpMsg.udpipmsg[sizeof(udpiphdr)];

void
fumiedInit()
{
  int o = 1;
  struct ifreq	ifr;
  memset(&ifr,0,sizeof(struct ifreq));
  memcpy(ifr.ifr_name,IfName,IfName_len);

  fumiedSocket = socket(AF_PACKET,SOCK_PACKET,htons(ETH_P_ALL));
  if ( fumiedSocket == -1 )
    {
      error("socket (fumiedInit)");
      exit(1);
    }
  if ( ioctl(fumiedSocket,SIOCGIFHWADDR,&ifr) )
    {
      error("ioctl SIOCGIFHWADDR (fumiedInit)");
      exit(1);
    }
  if ( ifr.ifr_hwaddr.sa_family != ARPHRD_ETHER )
    {
      error("interface is not Ethernet");
      exit(1);
    }
  if ( setsockopt(fumiedSocket,SOL_SOCKET,SO_BROADCAST,&o,sizeof(o)) == -1 )
    {
      error("setsockopt (fumiedInit)");
      exit(1);
    }
  ifr.ifr_flags = IFF_UP | IFF_BROADCAST | IFF_MULTICAST| IFF_NOTRAILERS | IFF_RUNNING;
  if ( ioctl(fumiedSocket,SIOCSIFFLAGS,&ifr) )
    {
      error("ioctl SIOCSIFFLAGS (fumiedInit)");
      exit(1);
    }
  memcpy(ClientHwAddr,ifr.ifr_hwaddr.sa_data,ETHER_ADDR_LEN);
  if ( ioctl(fumiedSocket, SIOCGIFADDR, &ifr) < 0 )
    {
      error("ioctl SIOCGIFADDR (fumiedInit)");
    }
  Ifbuf.addr = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr;
}

void
fumiedClient()
{
  int next_state;

  Fsm[INIT]       = init;
  Fsm[REQUESTING] = request;
  Fsm[GETTING]    = getinfo;

  CurrState = PrevState = INIT;

  for(;;)
    {
    next_state = (*Fsm[CurrState])();
    if ( next_state == EXCEPTION )
      {
	error("Exception occured in the fsm (fumiedClient)");
      }
    else if ( next_state == COMPLETED )
      {
	note("Request completed");
	return;
      }
    PrevState = CurrState;
    CurrState  = next_state;
    }
}

int
init()
{
  int waitTime = 0;
  struct sockaddr	addr;

  mkFumiedDiscoverMsg();
  memset(&addr,0,sizeof(struct sockaddr));
  memcpy(addr.sa_data,IfName,IfName_len);

  do
    {
      waitTime = random() % 10;
    }
  while ( waitTime == 0 );
  sleep(waitTime);
  time(&ReqSentTime);	/* record local time
			   not @ REQUEST but @ DISCOVER (RFC1541) */
  if ( sendto(fumiedSocket,&UdpIpMsg,sizeof(struct ether_header)+
	      sizeof(udpiphdr)+sizeof(struct fumiedMessage),0,
	      &addr,sizeof(struct sockaddr)) == -1 )
    {
      error("sendto (init)");
    }
  return REQUESTING;
}

int
request()
{
  struct sockaddr addr;
  char local_hostname[80];
  char local_ipaddr[80];
  struct utsname sysname;
  struct in_addr l_addr;

  memset(&addr,0,sizeof(struct sockaddr));
  memcpy(addr.sa_data,IfName,IfName_len);

  if ( waitChkReXmitMsg(&addr, N_REXMIT_REQUEST) )
    {
    mkFumiedRequestMsg();
#ifdef NEED_BCAST_RESPONSE
    FumiedMsgSend->flags = htons(F_BROADCAST);
#endif

  if ( sendto(fumiedSocket,&UdpIpMsg,sizeof(struct ether_header)+
	      sizeof(udpiphdr)+sizeof(struct fumiedMessage),0,
	      &addr,sizeof(struct sockaddr)) == -1 )
    {
    error("sendto (selecting)");
    }
    return GETTING;
    }
  else
    {
    l_addr.s_addr = Ifbuf.addr;
    strcpy(local_ipaddr, inet_ntoa(l_addr));
    if (!uname(&sysname))
      {
	strcpy(local_hostname, sysname.nodename);
      }
    note("local hostname:%s", local_hostname);
    note("local ipaddr:%s", local_ipaddr);
    revise_hosts(local_hostname, local_ipaddr, NULL, NULL);
    return COMPLETED;
    }
}

int
getinfo()
{
  struct sockaddr addr;
  char peer_hostname[80];
  char local_hostname[80];
  char peer_ipaddr[80];
  char local_ipaddr[80];
  struct utsname sysname;
  struct in_addr l_addr;

  memset(&addr,0,sizeof(struct sockaddr));
  memcpy(addr.sa_data,IfName,IfName_len);
   
  if ( waitChkReXmitMsg(&addr, N_REXMIT_REQUEST) )
    {
    strncpy(peer_hostname, (u_char *)(FumiedMsgRecv->options)+5, sizeof(peer_hostname)); 
    strcpy(peer_ipaddr, inet_ntoa(FumiedMsgRecv->siaddr));
    note("peer hostname:%s", peer_hostname);
    note("peer ipaddr:%s", peer_ipaddr);
    l_addr.s_addr = Ifbuf.addr;
    strcpy(local_ipaddr, inet_ntoa(l_addr));
    if (!uname(&sysname))
      {
	strcpy(local_hostname, sysname.nodename);
      }
    note("local hostname:%s", local_hostname);
    note("local ipaddr:%s", local_ipaddr);
    revise_hosts(local_hostname, local_ipaddr, peer_hostname,
		 peer_ipaddr);
    return COMPLETED;
  }
  warn("no response to FUMIEDREEQUEST message. move to INIT state");
  return EXCEPTION;
}

void
mkFumiedDiscoverMsg()
{
  register unsigned char *p = FumiedMsgSend->options + 4;

/* build Ethernet header */
  memset(&UdpIpMsg,0,sizeof(udpipMessage));
  memcpy(UdpIpMsg.ethhdr.ether_dhost,MAC_BCAST_ADDR,ETHER_ADDR_LEN);
  memcpy(UdpIpMsg.ethhdr.ether_shost,ClientHwAddr,ETHER_ADDR_LEN);
  UdpIpMsg.ethhdr.ether_type = htons(ETHERTYPE_IP);

  FumiedMsgSend->htype	=	ARPHRD_ETHER;
  FumiedMsgSend->hlen	=	ETHER_ADDR_LEN;
#ifdef NEED_BCAST_RESPONSE
  FumiedMsgSend->flags	=	htons(BROADCAST_FLAG);
#endif
  FumiedMsgSend->ciaddr.s_addr = Ifbuf.addr;
  *((unsigned long *)(FumiedMsgSend->options)) = htonl(MAGIC_COOKIE);
  *p++ = FUMIEDDISCOVER;

/* build UDP/IP header */
  udpipgen((udpiphdr *)UdpIpMsg.udpipmsg,0,INADDR_BROADCAST,
  htons(FUMIED_CLIENT_PORT),htons(FUMIED_SERVER_PORT),sizeof(struct fumiedMessage));
}

void
mkFumiedRequestMsg(struct fumiedMessage *msg)
{
  register unsigned char *p = FumiedMsgSend->options + 4;
  struct utsname sysname;

/* build Ethernet header */
  memset(&UdpIpMsg,0,sizeof(udpipMessage));
  memcpy(UdpIpMsg.ethhdr.ether_dhost,MAC_BCAST_ADDR,ETHER_ADDR_LEN);
  memcpy(UdpIpMsg.ethhdr.ether_shost,ClientHwAddr,ETHER_ADDR_LEN);
  UdpIpMsg.ethhdr.ether_type = htons(ETHERTYPE_IP);

  FumiedMsgSend->htype	=	ARPHRD_ETHER;
  FumiedMsgSend->hlen	=	ETHER_ADDR_LEN;
#ifdef NEED_BCAST_RESPONSE
  FumiedMsgSend->flags	=	htons(BROADCAST_FLAG);
#endif
  FumiedMsgSend->ciaddr.s_addr = Ifbuf.addr;
  *((unsigned long *)(FumiedMsgSend->options)) = htonl(MAGIC_COOKIE);
  *p++ = FUMIEDREQUEST;
  if (!uname(&sysname))
    {
      strcpy(p, sysname.nodename);
      p += strlen(sysname.nodename);
    }
  *p++ = '\0';

/* build UDP/IP header */
  udpipgen((udpiphdr *)UdpIpMsg.udpipmsg,0,INADDR_BROADCAST,
  htons(FUMIED_CLIENT_PORT),htons(FUMIED_SERVER_PORT),sizeof(struct fumiedMessage));
}

long
getNextTimeout(int flag)
{
  static long prevTimeout;
  long rv;			/* return value */

  if ( flag == INIT_TIMEOUT )
    {
      prevTimeout = 2;		/* 2 seconds */
    }
  else if ( prevTimeout > 64 )
    {
      prevTimeout = 2;
    }

  rv = (prevTimeout - 1) * 1000000 + rand()/2000000;
  prevTimeout *= 2;
  return rv;
}
